/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.io.Closeables;
import java.io.BufferedReader;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
public class CloseableIterators {
    private static final CloseableIterator EMPTY = new CloseableIterator(Collections.emptyIterator(), Closeables.noop());

    public static <T> CloseableIterator<T> empty() {
        return EMPTY;
    }

    public static CloseableIterator<String> lines(BufferedReader reader) throws IOException {
        final String firstLine = reader.readLine();
        if (firstLine == null)
            return additionallyClosing(CloseableIterators.<String>empty(), reader);
        else
            return new CloseableIterator(new LineBufferedReaderIterator(firstLine, reader), reader);
    }

    public static CloseableIterator<Character> characters(BufferedReader reader) throws IOException {
        int firstCharacter = reader.read();
        if (firstCharacter < 0)
            return additionallyClosing(CloseableIterators.<Character>empty(), reader);
        else
            return new CloseableIterator(new CharacterBufferedReaderIterator((char)firstCharacter, reader), reader);
    }

    public static CloseableIterator<Byte> bytes(InputStream inputStream) throws IOException {
        int firstByte = inputStream.read();
        if (firstByte < 0)
            return additionallyClosing(CloseableIterators.<Byte>empty(), inputStream);
        else
            return new CloseableIterator(new ByteInputStreamIterator((byte)firstByte, inputStream), inputStream);
    }

    public static <T> CloseableIterator<T> iterator(Iterator<T> iterator) {
        return new CloseableIterator<T>(iterator, Closeables.noop());
    }

    public static <T, U extends Iterator<T> & Closeable> CloseableIterator<T> of(final U value) {
        if (value instanceof CloseableIterators)
            return (CloseableIterator<T>)value;
        else
            return CloseableIterator.newInstance(value);
    }

    public static <T, U extends Iterator<T> & Closeable> CloseableIterator<T> withAdditionalClosingAction(final U value, final Runnable action) {
        return new CloseableIterator<T>(value, Closeables.combine(value, Closeables.runnable(action)));
    }

    public static <T, U extends Iterator<T> & Closeable> CloseableIterator<T> additionallyClosing(final U value, final Closeable closeable) {
        return new CloseableIterator<T>(value, Closeables.combine(value, closeable));
    }

    private static class LineBufferedReaderIterator implements Iterator<String> {
        private final BufferedReader bufferedReader;
        private boolean hasNext = true;
        private String nextLine;

        public LineBufferedReaderIterator(String firstLine, BufferedReader bufferedReader) {
            this.nextLine = firstLine;
            this.bufferedReader = bufferedReader;
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public String next() {
            if (!hasNext)
                throw new NoSuchElementException();
            String result = nextLine;
            try {
                nextLine = bufferedReader.readLine();
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            }
            hasNext = nextLine != null;
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Not supported.");
        }
    }

    private static class CharacterBufferedReaderIterator implements Iterator<Character> {
        private final BufferedReader reader;
        private char nextCharacter;
        private boolean hasNext = true;

        public CharacterBufferedReaderIterator(char firstCharacter, BufferedReader reader) {
            this.nextCharacter = firstCharacter;
            this.reader = reader;
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public Character next() {
            if (!hasNext)
                throw new NoSuchElementException();
            char result = nextCharacter;
            int character;
            try {
                character = reader.read();
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            }
            if (character < 0)
                hasNext = false;
            else
                nextCharacter = (char)character;
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Not supported.");
        }
    }

    private static class ByteInputStreamIterator implements Iterator<Byte> {
        private final InputStream inputStream;
        private byte nextByte;
        private boolean hasNext = true;

        public ByteInputStreamIterator(byte firstByte, InputStream inputStream) {
            this.nextByte = firstByte;
            this.inputStream = inputStream;
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public Byte next() {
            if (!hasNext)
                throw new NoSuchElementException();
            byte result = nextByte;
            int byteValue;
            try {
                byteValue = inputStream.read();
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            }
            if (byteValue < 0)
                hasNext = false;
            else
                nextByte = (byte)byteValue;
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Not supported.");
        }
    }
}
