/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.Consumer;
import com.github.sviperll.Predicate;
import java.io.Closeable;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
class FilteredStreamBehaviour<T> implements StreamBehaviour<T> {
    private final StreamBehaviour<T> upstream;
    private final Predicate<? super T> predicate;
    public FilteredStreamBehaviour(StreamBehaviour<T> upstream, Predicate<? super T> predicate) {
        this.upstream = upstream;
        this.predicate = predicate;
    }

    @Override
    public CloseableIterator<T> openCloseableIterator() {
        CloseableIterator<T> upstreamIterator = upstream.openCloseableIterator();
        try {
            if (!upstreamIterator.hasNext())
                return CloseableIterators.empty();
            else {
                T element = upstreamIterator.next();
                while (!predicate.evaluate(element)) {
                    if (!upstreamIterator.hasNext())
                        return CloseableIterators.empty();
                    else
                        element = upstreamIterator.next();
                }
                final T firstElement = element;
                CloseableIterator<T> result = CloseableIterators.of(new FilteredIterator(firstElement, upstreamIterator));
                upstreamIterator = null;
                return result;
            }
        } finally {
            if (upstreamIterator != null)
                upstreamIterator.close();
        }
    }

    @Override
    public void forEach(final Consumer<? super T> consumer) {
        upstream.forEach(new Consumer<T>() {
            @Override
            public void accept(T value) {
                if (predicate.evaluate(value))
                    consumer.accept(value);
            }
        });
    }

    private class FilteredIterator implements Iterator<T>, Closeable {
        private final CloseableIterator<T> upstreamIterator;
        private boolean hasNext = true;
        private T nextElement;

        public FilteredIterator(T firstElement,
                                CloseableIterator<T> downstreamIterator) {
            this.nextElement = firstElement;
            this.upstreamIterator = downstreamIterator;
        }

        @Override
        public void close() {
            upstreamIterator.close();
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public T next() {
            if (!hasNext)
                throw new NoSuchElementException();
            T result = nextElement;
            setNextElement();
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        private void setNextElement() {
            if (!upstreamIterator.hasNext()) {
                hasNext = false;
                nextElement = null;
            } else {
                T element = upstreamIterator.next();
                while (!predicate.evaluate(element)) {
                    if (!upstreamIterator.hasNext()) {
                        hasNext = false;
                        nextElement = null;
                        return;
                    } else
                        element = upstreamIterator.next();
                }
                hasNext = true;
                nextElement = element;
            }
        }
    }

}
