/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.Consumer;
import com.github.sviperll.Generator;
import java.io.Closeable;
import java.util.Iterator;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
class FlattenedStreamBehaviour<T> implements StreamBehaviour<T> {
    private final StreamBehaviour<? extends Generator<T>> stream;
    public FlattenedStreamBehaviour(StreamBehaviour<? extends Generator<T>> stream) {
        this.stream = stream;
    }

    @Override
    public CloseableIterator<T> openCloseableIterator() {
        CloseableIterator<? extends Generator<T>> outerIterator = stream.openCloseableIterator();
        try {
            if (!outerIterator.hasNext())
                return CloseableIterators.empty();
            else {
                CloseableIterator<T> iterator = Streams.generator(outerIterator.next()).openCloseableIterator();
                try {
                    while (!iterator.hasNext()) {
                        iterator.close();
                        iterator = null;
                        if (!outerIterator.hasNext())
                            return CloseableIterators.empty();
                        else
                            iterator = Streams.generator(outerIterator.next()).openCloseableIterator();
                    }
                    final T firstElement = iterator.next();
                    while (!iterator.hasNext()) {
                        iterator.close();
                        iterator = null;
                        if (!outerIterator.hasNext())
                            return CloseableIterators.empty();
                        else
                            iterator = Streams.generator(outerIterator.next()).openCloseableIterator();
                    }
                    CloseableIterator<T> result = CloseableIterators.of(new FlattenedIterator(iterator, firstElement, outerIterator));
                    iterator = null;
                    outerIterator = null;
                    return result;
                } finally {
                    if (iterator != null)
                        iterator.close();
                }
            }
        } finally {
            if (outerIterator != null)
                outerIterator.close();
        }
    }

    @Override
    public void forEach(final Consumer<? super T> consumer) {
        stream.forEach(new Consumer<Generator<T>>() {
            @Override
            public void accept(Generator<T> value) {
                value.forEach(consumer);
            }
        });
    }

    private class FlattenedIterator implements Iterator<T>, Closeable {

        private final CloseableIterator<? extends Generator<T>> enclosinIterator;
        private CloseableIterator<T> openedIterator;
        private T nextElement;
        private boolean hasNext = true;

        public FlattenedIterator(CloseableIterator<T> firstIterator,
                                          T firstElement,
                                          CloseableIterator<? extends Generator<T>> enclosinIterator) {
            this.openedIterator = firstIterator;
            this.nextElement = firstElement;
            this.enclosinIterator = enclosinIterator;
        }

        @Override
        public void close() {
            try {
                if (openedIterator != null)
                    openedIterator.close();
            } finally {
                enclosinIterator.close();
            }
        }

        @Override
        public boolean hasNext() {
            return hasNext;
        }

        @Override
        public T next() {
            T result = nextElement;
            while (!openedIterator.hasNext()) {
                openedIterator.close();
                openedIterator = null;
                if (!enclosinIterator.hasNext()) {
                    hasNext = false;
                    return result;
                } else
                    openedIterator = Streams.generator(enclosinIterator.next()).openCloseableIterator();
            }
            nextElement = openedIterator.next();
            return result;
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException("Not supported.");
        }
    }

}
