/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.Consumer;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
class LimitedStreamBehaviour<T> implements StreamBehaviour<T> {
    private final StreamBehaviour<T> stream;
    private final int size;

    public LimitedStreamBehaviour(StreamBehaviour<T> stream, int size) {
        this.stream = stream;
        this.size = size;
    }

    @Override
    public CloseableIterator<T> openCloseableIterator() {
        final CloseableIterator<T> iterator = stream.openCloseableIterator();
        Iterator<T> limitedIterator = new Iterator<T>() {
            private int count = 0;

            @Override
            public boolean hasNext() {
                return iterator.hasNext() && count < size;
            }

            @Override
            public T next() {
                if (!hasNext())
                    throw new NoSuchElementException();
                T result = iterator.next();
                count++;
                return result;
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException("Not supported.");
            }
        };
        return new CloseableIterator<T>(limitedIterator, iterator);
    }

    @Override
    public void forEach(final Consumer<? super T> consumer) {
        try {
            stream.forEach(new Consumer<T>() {
                int count = 0;
                @Override
                public void accept(T value) {
                    if (count < size)
                        consumer.accept(value);
                    else
                        throw new LimitExceededException();
                    count++;
                }
            });
        } catch (LimitExceededException ex) {
            // Normal exit
        }
    }

    private static class LimitExceededException extends RuntimeException {
    }

}
