/*
 * Copyright 2013 Victor Nazarov <asviraspossible@gmail.com>.
 */
package org.github.sviperll.stream;

import com.github.sviperll.BiFunction;
import com.github.sviperll.BinaryOperator;
import com.github.sviperll.Consumer;
import com.github.sviperll.Function;
import com.github.sviperll.Generator;
import com.github.sviperll.OptionalVisitor;
import com.github.sviperll.Predicate;
import com.github.sviperll.Supplier;
import com.github.sviperll.UnaryOperator;
import java.awt.Desktop;
import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.github.sviperll.stream.Collector.CollectorState;

/**
 *
 * @author Victor Nazarov <asviraspossible@gmail.com>
 */
public class Streams {
    public static <T> Stream<T> generator(final Generator<T> generator) {
        if (generator instanceof StreamBehaviour)
            return behaviour((StreamBehaviour<T>)generator);
        else
            return new Stream<T>(new GeneratorStreamBehaviour<T>(generator));
    }

    public static <T> Stream<T> iterable(final Iterable<T> iterable) {
        return new Stream<T>(new IterableStreamBehaviour<T>(iterable));
    }

    public static <T> Stream<T> array(final T[] array) {
        return new Stream<T>(new ArrayStreamBehaviour<T>(array));
    }

    public static Stream<String> lines(File file, Charset charset) {
        return new Stream<String>(new FileLinesStreamBehaviour(file, charset));
    }

    public static Stream<Character> characters(File file, Charset charset) {
        return new Stream<Character>(new FileCharactersStreamBehaviour(file, charset));
    }

    public static Stream<Byte> bytes(File file) {
        return new Stream<Byte>(new FileBytesStreamBehaviour(file));
    }

    private static <T> Stream<T> behaviour(StreamBehaviour<T> behaviour) {
        if (behaviour instanceof Stream)
            return (Stream<T>)behaviour;
        else
            return new Stream<T>(behaviour);
    }

    public static <T> Stream<T> generate(final Supplier<T> supplier) {
        return new Stream<T>(new GeneratedStreamBehaviour(supplier));
    }

    public static <T> Stream<T> iterate(final T seed, final UnaryOperator<T> function) {
        return new Stream<T>(new IteratedStreamBehaviour<T>(seed, function));
    }

    public static <T, U> Stream<U> map(StreamBehaviour<T> stream, Function<? super T, U> function) {
        return new Stream<U>(new MappedStreamBehaviour<T, U>(stream, function));
    }

    public static <T> Stream<T> filter(StreamBehaviour<T> stream, Predicate<? super T> predicate) {
        return new Stream<T>(new FilteredStreamBehaviour<T>(stream, predicate));
    }

    public static <T> Stream<T> limit(final StreamBehaviour<T> stream, final int size) {
        return new Stream<T>(new LimitedStreamBehaviour<T>(stream, size));
    }

    public static <T> Stream<T> skip(final StreamBehaviour<T> stream, final int offset) {
        return new Stream<T>(new SkippedStreamBehaviour<T>(stream, offset));
    }

    public static <T, R, E extends Exception> R collect(StreamBehaviour<T> stream, Supplier<? extends CollectorState<? super T, R, E>> collector) throws E {
        CollectorState<? super T, R, E> state = collector.get();
        stream.forEach(state);
        return state.result();
    }

    public static <T, U> U reduce(final StreamBehaviour<T> stream, final U seed, final BiFunction<U, ? super T, U> function) {
        return Streams.<T, U, RuntimeException>collect(stream, Collectors.reducing(seed, function));
    }

    public static <T, R, E extends Exception> R reduce(final StreamBehaviour<T> stream, BinaryOperator<T> operator, OptionalVisitor<? super T, R, E> visitor) throws E {
        return collect(stream, Collectors.reducing(operator, visitor));
    }

    public static <T> Stream<T> flatten(StreamBehaviour<? extends Generator<T>> stream) {
        return new Stream<T>(new FlattenedStreamBehaviour<T>(stream));
    }

    public static <T, R> Stream<R> flatMap(StreamBehaviour<T> stream, Function<? super T, ? extends Generator<R>> function) {
        return flatten(new MappedStreamBehaviour<T, Generator<R>>(stream, function));
    }

    public static <T, R, E extends Exception> R findFirst(final StreamBehaviour<T> stream, OptionalVisitor<? super T, R, E> visitor) throws E {
        return collect(limit(stream, 1), Collectors.findFirst(visitor));
    }

    public static <T> int count(final StreamBehaviour<T> stream) {
        return collect(stream, Collectors.counting());
    }

    public static <T> Stream<T> withAdditionalClosingAction(final StreamBehaviour<T> stream, final Runnable action) {
        return new Stream<T>(new StreamBehaviour<T>() {
            @Override
            public CloseableIterator<T> openCloseableIterator() {
                return CloseableIterators.withAdditionalClosingAction(stream.openCloseableIterator(), action);
            }

            @Override
            public void forEach(Consumer<? super T> consumer) {
                try {
                    stream.forEach(consumer);
                } finally {
                    action.run();
                }
            }
        });
    }

    public static <T> Stream<T> factory(final Supplier<? extends StreamBehaviour<T>> streamFactory) {
        return new Stream<T>(new StreamBehaviour<T>() {
            @Override
            public CloseableIterator<T> openCloseableIterator() {
                return streamFactory.get().openCloseableIterator();
            }

            @Override
            public void forEach(Consumer<? super T> consumer) {
                streamFactory.get().forEach(consumer);
            }
        });
    }

    public static <T, U extends Closeable & Iterator<T>> Stream<T> iteratorFactory(final Supplier<U> iteratorFactory) {
        return new Stream<T>(new StreamBehaviour<T>() {
            @Override
            public CloseableIterator<T> openCloseableIterator() {
                return CloseableIterators.of(iteratorFactory.get());
            }

            @Override
            public void forEach(Consumer<? super T> consumer) {
                U iterator = iteratorFactory.get();
                try {
                    while (iterator.hasNext()) {
                        consumer.accept(iterator.next());
                    }
                } finally {
                    try {
                        iterator.close();
                    } catch (IOException ex) {
                        throw new RuntimeException(ex);
                    }
                }
            }
        });
    }
}
